#include <stdio.h>
#include <string.h>
#include <conio.h>

typedef unsigned char  uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long  uint32_t;

uint16_t mpu_base = 0x330;
uint16_t sb_base = 0x220;

uint16_t dump_addr;

uint8_t dsp_ver_h;
uint8_t dsp_ver_l;

uint32_t crc6k;
uint32_t crc8k;

uint16_t error = 0;

#define SB_RESET_PORT   (sb_base + 0x6)
#define SB_READ_DATA    (sb_base + 0xA)
#define SB_WRITE_DATA   (sb_base + 0xC)
#define SB_WRITE_STATUS (sb_base + 0xC)
#define SB_READ_STATUS  (sb_base + 0xE)

#define ERROR_SB_RESET        (1)
#define ERROR_SB_SEND         (2)
#define ERROR_SB_RECV         (3)
#define ERROR_MPU_SEND_CMD    (4)
#define ERROR_MPU_SEND_DATA   (5)
#define ERROR_MPU_RECV_DATA   (6)
#define ERROR_MPU_LOOPBACK    (7)
#define ERROR_DSP_UNKNOWN     (8)
#define ERROR_DUMP_FAILED     (9)

char strbuf[256];
uint8_t memdump[256];

uint8_t dump[8192];

FILE *logfile;

void log_open (void) {
  logfile = fopen ("log.txt", "at");
}

void log_close (void) {
  if (logfile != NULL) fclose (logfile);
}

void log (const char *str) {
  printf ("%s", str);
  if (logfile != NULL) fputs (str, logfile);
}



static uint8_t mpu_output_full (void) {
  return(inp(mpu_base + 1) & 0x40);
}
static unsigned char mpu_input_empty(void) {
  return(inp(mpu_base + 1) & 0x80);
}
static uint8_t mpu_read (void) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_input_empty() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_RECV_DATA; return 0; }
  return inp(mpu_base);
}
static void mpu_write (uint8_t byte) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_output_full() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_SEND_DATA; return; }
  outp (mpu_base, byte);
}
static void mpu_command (uint8_t cmd) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (mpu_output_full() == 0) break;
  }
  if (i == 0xFFFF) { error = ERROR_MPU_SEND_CMD; return; }
  outp (mpu_base + 1, cmd);
}


void mpu_reset (void) {
  mpu_command (0xFF);
  mpu_read();
}

void mpu_init_try (void) {
  uint8_t b;
  mpu_command (0xFF);
  if (error) return;
  b = mpu_read();
  if (error) return;
  if (b != 0xFE) {
    error = ERROR_MPU_RECV_DATA;
    return;
  }
  mpu_command (0x3F);
  if (error) return;
  b = mpu_read();
  if (error) return;
  if (b != 0xFE) error = ERROR_MPU_RECV_DATA;

}


void mpu_init (void) {
  error = 0;
  mpu_init_try ();
  if (error == 0) return;

  for (;;) {
    mpu_command (0xFF);
    sb_delay (0xFFFF);
    inp (SB_READ_DATA);
    inp (mpu_base);
    error = 0;
    sb_reset ();
    mpu_init_try ();
    if (error == 0) return;
  }

}

void mpu_loopback_check (void) {
  mpu_write (0xCA);
  if (error) return;
  if (mpu_read() != 0xCA) error = ERROR_MPU_LOOPBACK;
  if (error) return;
  mpu_write (0xFE);
  if (error) return;
  if (mpu_read() != 0xFE) error = ERROR_MPU_LOOPBACK;
  if (error) return;
}


static uint8_t sb_read_data_available (void) {
  return inp(SB_READ_STATUS) & 0x80;
}

static void sb_delay (uint16_t cycles) {  // 1 cycle is at least 480ns
  uint16_t i;
  for (i=0; i<cycles; ++i) inp(SB_WRITE_STATUS);
}

static void sb_send (uint8_t data) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if ((inp(SB_WRITE_STATUS) & 0x80) == 0x00) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_SEND; return; }

  outp (SB_WRITE_DATA, data);

}

static uint8_t sb_recv (void) {
  uint16_t i;

  for (i=0; i < 0xFFFF; ++i) {
    if (inp(SB_READ_STATUS) & 0x80) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_RECV; return 0; }

  return inp (SB_READ_DATA);
}


void sb_reset (void) {
  uint16_t i;

  outp (SB_RESET_PORT, 1);
  inp(SB_READ_DATA);
  sb_delay (7);  // more then 3us
  outp (SB_RESET_PORT, 0);
  for (i=0; i < 0xFFFF; ++i) { // 208 = ~100us
    if (sb_read_data_available()) break;
  }
  if (i == 0xFFFF) { error = ERROR_SB_RESET; return; }

  if (inp(SB_READ_DATA) != 0xAA) error = ERROR_SB_RESET;
}

void sb_init (void) {
  inp (SB_READ_DATA);
  inp (mpu_base);
  mpu_command (0xFF);
}



void sb_get_dsp_ver (void) {
  uint8_t h,l;

  sb_send (0xE1);
  if (error) return;
  dsp_ver_h = sb_recv ();
  if (error) return;
  dsp_ver_l = sb_recv ();
  if (error) return;

}

static uint8_t sb_read_memory (uint8_t addr) {
  sb_send (0xF9);
  if (error) return 0;
  sb_send (addr);
  if (error) return 0;
  return sb_recv ();
}

static void sb_write_memory (uint8_t addr, uint8_t data) {
  sb_send (0xFA);
  if (error) return;
  sb_send (addr);
  if (error) return;
  sb_send (data);
}

void sb_clean_memory (uint8_t data) {
  int addr;

  for (addr = 0x40; addr < 256; ++addr) {
    sb_write_memory (addr, data);
    if (error) return;
  }
}

void sb_dump_memory (uint8_t *ptr) {
  int addr;

  for (addr = 0; addr < 256; ++addr) {
    ptr[addr] = sb_read_memory (addr);
    if (error) return;
  }
}

void sb_save_memdump (uint8_t *ptr, char *filename) {
  FILE *file;
  file = fopen (filename, "wb");
  if (file != NULL) {
    fwrite (ptr, 256, 1, file);
    fclose (file);
  }
}

void memory_dump (void) {
  sb_clean_memory (0xFF);
  if (error) return;
  sb_dump_memory (memdump);
  if (error) return;
  sb_save_memdump (memdump, "MDUMP2.bin");
  sb_clean_memory (0x00);
  if (error) return;
  sb_dump_memory (memdump);
  if (error) return;
  sb_save_memdump (memdump, "MDUMP1.bin");

}

void sb_initial_diag (void) {
  error = 0;
  log ("SB Reset: ");
  sb_reset ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  log ("DSP version: ");
  sb_get_dsp_ver ();
  if (error) { log ("failed!\n"); return; }
  else {
    sprintf (strbuf, "%d.%02d\n", (int)dsp_ver_h, (int)dsp_ver_l);
    log (strbuf);
  }

  if (dsp_ver_h != 0x04) { error = ERROR_DSP_UNKNOWN; return; }


  log ("MPU-401 init: ");
  mpu_init ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  log ("Internal memory dump: ");
  memory_dump ();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");


  log ("MPU-401 loopback check: ");
  mpu_loopback_check();
  if (error) { log ("failed!\n"); return; }
  else log ("done\n");

  switch (dsp_ver_l) {
    case 4:
    break;
    case 5:
    break;
    case 11:
    break;
    case 13:
    break;
    case 16:
    break;
    default:
    error = ERROR_DSP_UNKNOWN;
    return;
  }

}

void save_dump (void);

/*
void dsp_413_attack1 (uint16_t addr) {
  error = 0;
  sb_reset ();
  mpu_init ();

  sb_write_memory (0x01, 0xC4);

  sb_write_memory (0xBE, 0x04);
  sb_write_memory (0xBD, 0x18);

  sb_write_memory (0xBC, addr >> 8);
  sb_write_memory (0xBB, addr & 0xFF);

  sb_write_memory (0xB9, 0x11);
  sb_write_memory (0xB8, 0xE9);

  mpu_write (0x04);

  dump[addr] = sb_recv ();
  dump[addr + 1] = sb_recv ();
}
*/

/*
void dsp_413_attack2 (uint16_t addr) {
  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0xBE, addr >> 8);
  sb_write_memory (0xBD, addr & 0xFF);

  sb_write_memory (0xBB, 0x11);
  sb_write_memory (0xBA, 0xE9);

  sb_write_memory (0x01, 0xC2);

  mpu_write (0x00);

  sb_delay (0xFFFF);

  sb_send (0xE1);

  dump[addr] = sb_recv ();
  dump[addr + 1] = sb_recv ();
}
*/

void dsp_404_attack1 (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0x7F, dump_addr >> 8);
  sb_write_memory (0x7E, dump_addr & 0xFF);

  sb_write_memory (0x7C, 0x0F);
  sb_write_memory (0x7B, 0x97);

  sb_write_memory (0x01, 0x82);

  mpu_write (0x02);

  sb_delay (0xFFFF);

  sb_send (0xE3);

  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}

void dsp_405_attack1 (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0x7F, dump_addr >> 8);
  sb_write_memory (0x7E, dump_addr & 0xFF);

  sb_write_memory (0x7C, 0x10);
  sb_write_memory (0x7B, 0x43);

  sb_write_memory (0x01, 0x82);

  mpu_write (0x02);

  sb_delay (0xFFFF);

  sb_send (0xE3);

  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}

void dsp_411_attack1 (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0xBE, dump_addr >> 8);
  sb_write_memory (0xBD, dump_addr & 0xFF);

  sb_write_memory (0xBB, 0x12);
  sb_write_memory (0xBA, 0x26);

  sb_write_memory (0x01, 0xC2);

  mpu_write (0x04);

  sb_delay (0xFFFF);

  sb_send (0xE3);

  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}


void dsp_413_attack3 (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();


  sb_write_memory (0xBE, dump_addr >> 8);
  sb_write_memory (0xBD, dump_addr & 0xFF);

  sb_write_memory (0xBB, 0x12);
  sb_write_memory (0xBA, 0x04);

  sb_write_memory (0x01, 0xC2);

  mpu_write (0x00);

  sb_delay (0xFFFF);
  inp (SB_READ_DATA);

  sb_send (0xE3);

  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    if (error) return;
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}

void dsp_416_attack (void) {
  uint8_t byte;
  int i;

  error = 0;
  sb_reset ();
  mpu_init ();

//  sb_write_memory (0xC0, dump_addr >> 8);
//  sb_write_memory (0xBF, dump_addr & 0xFF);

  // orig
  sb_write_memory (0xBF, dump_addr >> 8);
  sb_write_memory (0xBE, dump_addr & 0xFF);



  sb_write_memory (0xBC, 0x10);
  sb_write_memory (0xBB, 0xDC);

  sb_write_memory (0x01, 0xC2);

  mpu_write ( 0x03);
  sb_delay (0xFFFF);
  sb_send (0xE3);


  for (i=0; i<256; ++i) {
    byte = sb_recv ();
    if (error) return;
    dump[dump_addr++] = byte;
    if (dump_addr >= 8192) return;
    if (byte == 0x00) return;
  }
}



/*
void dsp_413_dump (void) {
  
  uint16_t i;

  dump[0] = 0;
  dsp_413_attack2 (0x0000);
  if (error) { error = ERROR_DUMP_FAILED; return; }
  if (dump[0] != 0x02) { error = ERROR_DUMP_FAILED; return; }

  for (i=2; i<8192; i+=2) {
    dsp_413_attack2 (i);
    if (error) { error = ERROR_DUMP_FAILED; return; }
    if ((i & 0xFF) == 0x00) {
      printf ("%04X dumped\n", i);
      dsp_save_dump ();
    }
  }
  
  sb_reset ();
  log ("Dump complete !\n");
  dsp_save_dump ();

}
*/

void do_attack (void) {
  switch (dsp_ver_l) {
    case 4:
    dsp_404_attack1 ();
    break;
    case 5:
    dsp_405_attack1 ();
    break;
    case 11:
    dsp_411_attack1 ();
    break;
    case 13:
    dsp_413_attack3 ();
    break;
    case 16:
    dsp_416_attack ();
    break;
    default:
    error = ERROR_DUMP_FAILED;
  }


}


void dsp_dump (void) {
  
  uint16_t i;
  uint16_t threshold;

  dump_addr = 0x0000;
  dump[0] = 0x00;
  do_attack ();
  if (error) { error = ERROR_DUMP_FAILED; return; }
  if (dump[0] != 0x02) { error = ERROR_DUMP_FAILED; return; }

  threshold = 0x0100;

  for (;;) {
    if (dump_addr >= 8192) break;
    do_attack ();
    if (dump_addr >= threshold) {
      printf ("%04X dumped\n", threshold);
      threshold += 0x100;
    }
  }
  
  sb_reset ();
  log ("Dump complete !\n");
  save_dump ();

}


/*
void dsp_413_dump (void) {
  
  uint16_t i;

  log ("Attack started\n");
  dump[0] = 0;
  dsp_413_attack2 (0x0000);
//  if (dump[0] != 0x02) {
//    error = ERROR_DUMP_FAILED;
//    return;
//    }
//  else log ("Attack successful\n");
  sb_reset ();
  dsp_save_dump ();
}
*/

void error_print (void) {
  if (error) log ("ERROR: ");

  switch (error) {
    case ERROR_SB_RESET:
    log ("Cannot RESET sound blaster\n");
    break;
    case ERROR_SB_SEND:
    log ("Cannot SEND to sound blaster\n");
    break;
    case ERROR_SB_RECV:
    log ("Cannot RECV from sound blaster\n");
    break;
    case ERROR_DSP_UNKNOWN:
    log ("DSP version not supported\n");
    break;

    case ERROR_MPU_SEND_CMD :
    log ("Cannot SEND cmd to MPU\n");
    break;
    case ERROR_MPU_SEND_DATA:
    log ("Cannot SEND data to MPU\n");
    break;
    case ERROR_MPU_RECV_DATA:
    log ("Cannot RECV from MPU\n");
    break;
    case ERROR_MPU_LOOPBACK :
    log ("Midi loopback not present\n");
    break;

    case ERROR_DUMP_FAILED:
    log ("Dump failed\n");
    break;


    default:
    log ("Success!\n");

  }

}

uint32_t crc32i_table[256];

void make_crc32i_table (void) {
  // (0x82608edb; 0x104c11db7) <=> (0xedb88320; 0x1db710641) {4294967263,91607,2974,268,171,91,57,34,21,12,10,10,10} | gold | (*p) IEEE 802.3; CRC-32
  // # 0x82608edb = x^32 +x^26 +x^23 +x^22 +x^16 +x^12 +x^11 +x^10 +x^8 +x^7 +x^5 +x^4 +x^2 +x +1 (0x82608edb; 0x104c11db7) <=> (0xedb88320; 0x1db710641)
  // HD6 for 268 bits = 33.5 bytes

  uint32_t reflected_poly = 0xedb88320;
  uint16_t j;
  uint8_t i;
  uint32_t crc;
  uint8_t bit;
  
  for (j = 0; j < 256; ++j) {
    crc = j;
    
    for (i = 0; i < 8; ++i) {
      bit = crc & 0x01;
      crc >>= 1;
      if (bit != 0) crc ^= reflected_poly;
    }
    
    crc32i_table[j] = crc;
  }

}

uint32_t calculate_crc32i (const uint8_t *data, uint16_t length) {
  uint32_t crc;
  crc = 0xFFFFFFFFl;
  while (length--)
    crc = crc32i_table[(crc ^ *data++) & 0xFFL] ^ (crc >> 8);
  return (crc ^ 0xFFFFFFFFl);
}


void test_dump_load (void) {
  FILE *file;
  file = fopen ("dump.bin", "rb");
  fread (dump, 8192, 1, file);
  fclose (file);
}

void save_dump_discrete (void) {
  FILE *file;

  log             ("DSP type  : discrete\n");
  sprintf (strbuf, "CRC32 (6k): %08lX (known)\n", crc6k);
  log (strbuf);
  sprintf (strbuf, "CRC32 (8k): %08lX (known)\n", crc8k);
  log (strbuf);

  sprintf (strbuf, "DSP_%d%02dD.bin", (int)dsp_ver_h, (int)dsp_ver_l);
  file = fopen (strbuf, "wb");
  if (file != NULL) {
    fwrite (dump, 8192, 1, file);
    fclose (file);
    log ("Saved     : ");
    log (strbuf);
    log ("\n");
  }
}

void save_dump_integrated (void) {
  FILE *file;

  sprintf (strbuf, "DSP type  : integrated (padding %02X)\n", (int)dump[0x1800]);
  log (strbuf);
  sprintf (strbuf, "CRC32 (6k): %08lX (known)\n", crc6k);
  log (strbuf);

  sprintf (strbuf, "DSP_%d%02dI.bin", (int)dsp_ver_h, (int)dsp_ver_l);
  file = fopen (strbuf, "wb");
  if (file != NULL) {
    fwrite (dump, 6*1024, 1, file);
    fclose (file);
    log ("Saved     : ");
    log (strbuf);
    log ("\n");
  }
}

void save_dump_unknown (void) {
  FILE *file;

  log             ("DSP type  : unknown\n");
  sprintf (strbuf, "CRC32 (6k): %08lX (unknown)\n", crc6k);
  log (strbuf);
  sprintf (strbuf, "CRC32 (8k): %08lX (unknown)\n", crc8k);
  log (strbuf);

  sprintf (strbuf, "%08lX.%d%02d", crc6k, (int)dsp_ver_h, (int)dsp_ver_l);
  file = fopen (strbuf, "wb");
  if (file != NULL) {
    fwrite (dump, 8192, 1, file);
    fclose (file);
    log ("Saved     : ");
    log (strbuf);
    log ("\n");
  }
}


uint8_t classify_dump (void) {

  make_crc32i_table ();

  crc6k = calculate_crc32i(dump, 6*1024);
  crc8k = calculate_crc32i(dump, 8*1024);

  switch (dsp_ver_l) {
    case 4:
    if (crc8k == 0x49FC3869l) return 1;
    else return 0;
    case 5:
    if (crc8k == 0xE51AFF23l) return 1;
    else return 0;

    case 11:
    if (crc6k != 0xC0533EE6l) return 0;
    if (crc8k == 0x4D75E821l) return 1;
    else return 2;


    case 13:
    if (crc6k != 0x9E1B22C6l) return 0;
    if (crc8k == 0xE22E9001l) return 1;
    else return 2;
    
    case 16:
    return 0;

    default:
    return 0;
  }

  return 0;
}

void save_dump (void) {
  uint8_t dump_type;

  dump_type = classify_dump ();
  switch (dump_type) {
    case 1:
    save_dump_discrete ();
    break;
    case 2:
    save_dump_integrated ();
    break;
    default:
    save_dump_unknown ();
    break;
  }

}

void main (int argc, char *argv[]) {
  

  error = 0;
  log_open ();
  sb_initial_diag ();

  if (error == 0) {
    dsp_dump ();
  }

  sb_reset ();

  error_print ();
  log_close ();
}
